/*----- PROTECTED REGION ID(E2Sextupole.h) ENABLED START -----*/
/* clang-format on */
//=============================================================================
//
// file :        E2Sextupole.h
//
// description : Include file for the E2Sextupole class
//
// project :     Elettra 2.0 pure corrector magnet device
//
// This file is part of Tango device class.
//
// Tango is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// Tango is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Tango.  If not, see <http://www.gnu.org/licenses/>.
//
//
// Elettra - Sincrotrone Trieste S.c.p.A.
//
//=============================================================================
//                This file is generated by POGO
//        (Program Obviously used to Generate tango Object)
//=============================================================================


#ifndef E2Sextupole_H
#define E2Sextupole_H

#include <tango/tango.h>
#include <PureSextupole.h>
#include <Corrector.h>
#include <Quadrupole.h>
/* clang-format off */
/*----- PROTECTED REGION END -----*/	//	E2Sextupole.h

#ifdef TANGO_LOG
	// cppTango after c934adea (Merge branch 'remove-cout-definition' into 'main', 2022-05-23)
	// nothing to do
#else
	// cppTango 9.3-backports and older
	#define TANGO_LOG       cout
	#define TANGO_LOG_INFO  cout2
	#define TANGO_LOG_DEBUG cout3
#endif // TANGO_LOG

/**
 *  E2Sextupole class description:
 *    Elettra 2.0 sextupole magnet device
 *    Use calibration table handle current/strgenght converstions.
 *    The magnet device is connected to a power supply tango device , real or simulated.
 *    
 *    Based on ESRF EBSmagnet concepts and libraries.
 */


namespace E2Sextupole_ns
{
/*----- PROTECTED REGION ID(E2Sextupole::Additional Class Declarations) ENABLED START -----*/
/* clang-format on */
//	Additional Class Declarations
const double default_beam_energy = 2.4e9; // default electron  beam energy in GeV
const double E0 =  510998.94999999995; // electron rest mass in eV
const double c_light = 299792458.0; // speed of light  m/s
//class Corrector;
class MainCurrentEventCallback;
/* clang-format off */
/*----- PROTECTED REGION END -----*/	//	E2Sextupole::Additional Class Declarations

class E2Sextupole : public TANGO_BASE_CLASS
{

/*----- PROTECTED REGION ID(E2Sextupole::Data Members) ENABLED START -----*/
/* clang-format on */
//	Add your own data members
	MagnetModel::PureSextupole* sextupole;  //magnet model
	MagnetModel::Corrector* ceh; //magnet model for embedded horizontal corrector
	MagnetModel::Corrector* cev; //magnet model for embedded vertical corrector
	MagnetModel::Quadrupole* skewquadrupole; // magnet model for embedded skew quadrupole
	double magnet_rigidity;
	double magnet_rigidity_inv;
	Tango::DeviceProxy *main_current_dev;
	Tango::DeviceProxy *ceh_current_dev;
	Tango::DeviceProxy *cev_current_dev;

	//main current
	MainCurrentEventCallback *main_current_cb;
	int main_current_event_id;
	int main_current_set_event_id;
	int main_state_event_id;
	Tango::WAttribute *mainCursetAtt;
	Tango::WAttribute *mainStrengthAtt;
	Tango::WAttribute *mainCorrectionAtt;
	Tango::WAttribute *mainResonanceAtt;
	std::vector<double> main_set_current; //the actual set point of the current (from the power supply)  used for strength calculations
	std::vector<double> main_set_strength; //the strength set point


	//ceh current - also QS
	MainCurrentEventCallback *ceh_current_cb;
	int ceh_current_event_id;
	int ceh_current_set_event_id;
	int ceh_state_event_id;
	Tango::WAttribute *cehCursetAtt;
	Tango::WAttribute *cehStrengthAtt;
	Tango::WAttribute *cehCorrectionAtt;
	Tango::WAttribute *cehResonanceAtt;
	std::vector<double> ceh_set_current; //the actual set point of the current (from the power supply)  used for strength calculations
	std::vector<double> ceh_set_strength; //the strength set point
	Tango::DevDouble ceh_design_strength; // may be later promoted to meomorized attribute


	//cev current
	MainCurrentEventCallback *cev_current_cb;
	int cev_current_event_id;
	int cev_current_set_event_id;
	int cev_state_event_id;
	Tango::WAttribute *cevCursetAtt;
	Tango::WAttribute *cevStrengthAtt;
	Tango::WAttribute *cevCorrectionAtt;
	Tango::WAttribute *cevResonanceAtt;
	std::vector<double> cev_set_current; //the actual set point of the current (from the power supply)  used for strength calculations
	std::vector<double> cev_set_strength; //the strength set point
	Tango::DevDouble cev_design_strength; // may be later promoted to meomorized attribute

     // add support for quadrupole

	bool design_strength_initialized; //true when design strenght has been called at least one , first call is to init valude from DB
public:
	omni_mutex iomutex; //guard for event and write interference
	Tango::DevState main_current_state;   // state of main current power supply device
	Tango::DevState ceh_current_state; // state of ceh current power supply device
	Tango::DevState cev_current_state; // state of cev current power supply device

	// main current
	Tango::DevDouble wValMainCorrStrength; //last CorrectionStrength write value
	Tango::DevDouble wValMainResStrength;  //last ResonanceStrength write value

	// ceh current
	Tango::DevDouble wValCehCorrStrength; //last CorrectionStrength write value
	Tango::DevDouble wValCehResStrength;  //last ResonanceStrength write value

	// cev current
	Tango::DevDouble wValCevCorrStrength; //last CorrectionStrength write value
	Tango::DevDouble wValCevResStrength;  //last ResonanceStrength write value
/* clang-format off */
/*----- PROTECTED REGION END -----*/	//	E2Sextupole::Data Members

//	Device property data members
public:
	//	CurrentDevice:	power supply tango device name
	std::string	currentDevice;
	//	Model:	corrector model (part number)
	std::string	model;
	//	SerialNumber:	serial number of the magnet
	std::string	serialNumber;
	//	MagnetName:	Name of magnet
	std::string	magnetName;
	//	CalibrationFileName:	name of calibration table file name
	std::string	calibrationFileName;
	//	CalibrationPath:	path of calibration and parameter files
	std::string	calibrationPath;
	//	BBASwitchable:	true if the magnet embbeded skew quadrupole cna be activated for beama based alignement operations.
	Tango::DevBoolean	bBASwitchable;
	//	SkewQuad:	True if the magnet is cabled with embbedd skew quadrupole QS . In this configuration the emmbedd horizontal corrector CEH is not available.
	Tango::DevBoolean	skewQuad;
	//	CEHName:	Magnet name of embbedded horizontal corrector.
	std::string	cEHName;
	//	CEVName:	Magnet name of embbedded vertical corrector.
	std::string	cEVName;
	//	CEHDeviceName:	Name of tango device implementing the interface of the embedded horizontal corrector.
	std::string	cEHDeviceName;
	//	CEVDeviceName:	Name of tango device implementing the interface of the embedded vertical corrector.
	std::string	cEVDeviceName;
	//	PsCehDeviceName:	Name of the tango device of powering the embbedded horizontal corrector / embedded skew quadrupole.
	std::string	psCehDeviceName;
	//	PsCevDeviceName:	Name of the tango device of powering the embbedded vertical corrector.
	std::string	psCevDeviceName;
	//	ParameterFileName:	name of auxiliary parametr file name
	//  (EBSmagnet compatibility)
	std::string	parameterFileName;
	//	CEHCalibrationFileName:	name of calibration table file name for horizontal embedded corrector - to be review3d when real calibrtaion tablles wil be available
	std::string	cEHCalibrationFileName;
	//	SQCalibrationFileName:	name of calibration table file name for embedded skeq quadrupole - to be reviewed when real calibration tables will be available
	std::string	sQCalibrationFileName;
	//	CEVCalibrationFileName:	name of calibration table file name for vertical embedded corrector - to be reviewed when real calibration tables wil be available
	std::string	cEVCalibrationFileName;

	bool	mandatoryNotDefined;

//	Attribute data members
public:
	Tango::DevDouble	*attr_Strength_read;
	Tango::DevDouble	*attr_CorrectionStrength_read;
	Tango::DevDouble	*attr_ResonanceStrength_read;
	Tango::DevDouble	*attr_Current_read;
	Tango::DevDouble	*attr_BeamEnergy_read;
	Tango::DevDouble	*attr_DesignStrength_read;
	Tango::DevBoolean	*attr_bbaEnable_read;
	Tango::DevDouble	*attr_CEHStrength_read;
	Tango::DevDouble	*attr_CEHCorrectionStrength_read;
	Tango::DevDouble	*attr_CEHResonanceStrength_read;
	Tango::DevDouble	*attr_CEVStrength_read;
	Tango::DevDouble	*attr_CEVCorrectionStrength_read;
	Tango::DevDouble	*attr_CEVResonanceStrength_read;
	Tango::DevDouble	*attr_SQStrength_read;
	Tango::DevDouble	*attr_SQCorrectionStrength_read;
	Tango::DevDouble	*attr_SQResonanceStrength_read;
	Tango::DevDouble	*attr_CEHCurrent_read;
	Tango::DevDouble	*attr_CEVCurrent_read;
	Tango::DevDouble	*attr_PolynomA_read;
	Tango::DevDouble	*attr_PolynomB_read;

//	Constructors and destructors
public:
	/**
	 * Constructs a newly device object.
	 *
	 *	@param cl	Class.
	 *	@param s 	Device Name
	 */
	E2Sextupole(Tango::DeviceClass *cl,std::string &s);
	/**
	 * Constructs a newly device object.
	 *
	 *	@param cl	Class.
	 *	@param s 	Device Name
	 */
	E2Sextupole(Tango::DeviceClass *cl,const char *s);
	/**
	 * Constructs a newly device object.
	 *
	 *	@param cl	Class.
	 *	@param s 	Device name
	 *	@param d	Device description.
	 */
	E2Sextupole(Tango::DeviceClass *cl,const char *s,const char *d);
	/**
	 * The device object destructor.
	 */
	~E2Sextupole();


//	Miscellaneous methods
public:
	/*
	 *	will be called at device destruction or at init command.
	 */
	void delete_device();
	/*
	 *	Initialize the device
	 */
	virtual void init_device();
	/*
	 *	Read the device properties from database
	 */
	void get_device_property();
	/*
	 *	Always executed method before execution command method.
	 */
	virtual void always_executed_hook();

	/*
	 *	Check if mandatory property has been set
	 */
	 void check_mandatory_property(Tango::DbDatum &class_prop, Tango::DbDatum &dev_prop);

//	Attribute methods
public:
	//--------------------------------------------------------
	/*
	 *	Method     : E2Sextupole::read_attr_hardware()
	 *	Description: Hardware acquisition for attributes.
	 */
	//--------------------------------------------------------
	virtual void read_attr_hardware(std::vector<long> &attr_list);
	//--------------------------------------------------------
	/*
	 *	Method     : E2Sextupole::write_attr_hardware()
	 *	Description: Hardware writing for attributes.
	 */
	//--------------------------------------------------------
	virtual void write_attr_hardware(std::vector<long> &attr_list);

/**
 *	Attribute Strength related methods
 *	Description: sextupole strength 1/m^3 ( m-3)
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
	virtual void read_Strength(Tango::Attribute &attr);
	virtual void write_Strength(Tango::WAttribute &attr);
	virtual bool is_Strength_allowed(Tango::AttReqType type);
/**
 *	Attribute CorrectionStrength related methods
 *
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
	virtual void read_CorrectionStrength(Tango::Attribute &attr);
	virtual void write_CorrectionStrength(Tango::WAttribute &attr);
	virtual bool is_CorrectionStrength_allowed(Tango::AttReqType type);
/**
 *	Attribute ResonanceStrength related methods
 *
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
	virtual void read_ResonanceStrength(Tango::Attribute &attr);
	virtual void write_ResonanceStrength(Tango::WAttribute &attr);
	virtual bool is_ResonanceStrength_allowed(Tango::AttReqType type);
/**
 *	Attribute Current related methods
 *
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
	virtual void read_Current(Tango::Attribute &attr);
	virtual void write_Current(Tango::WAttribute &attr);
	virtual bool is_Current_allowed(Tango::AttReqType type);
/**
 *	Attribute BeamEnergy related methods
 *	Description: Electron beam energy in eV. Used to calculate magnetic rigidity in strength calculations.
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
	virtual void read_BeamEnergy(Tango::Attribute &attr);
	virtual void write_BeamEnergy(Tango::WAttribute &attr);
	virtual bool is_BeamEnergy_allowed(Tango::AttReqType type);
/**
 *	Attribute DesignStrength related methods
 *	Description: sextupole design strength 1/m^3 ( m-3). The strenght reqiured by the nominal optic
 *               The attribute is memorized , expert only.
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
	virtual void read_DesignStrength(Tango::Attribute &attr);
	virtual void write_DesignStrength(Tango::WAttribute &attr);
	virtual bool is_DesignStrength_allowed(Tango::AttReqType type);
/**
 *	Attribute bbaEnable related methods
 *	Description: Read / and set the rele that modfy the sestupole cabling. If true the skew quadrupole QS is enabled, If fals the embedded horizontal corrector CEH is enabled.
 *               If the magnet is permanetly cabled with the embedded skew quadrupole, the aiitrbute is always true nd is not writable.
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Scalar
 */
	virtual void read_bbaEnable(Tango::Attribute &attr);
	virtual void write_bbaEnable(Tango::WAttribute &attr);
	virtual bool is_bbaEnable_allowed(Tango::AttReqType type);
/**
 *	Attribute CEHStrength related methods
 *	Description: embbeded corrector horizontal kick
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
	virtual void read_CEHStrength(Tango::Attribute &attr);
	virtual void write_CEHStrength(Tango::WAttribute &attr);
	virtual bool is_CEHStrength_allowed(Tango::AttReqType type);
/**
 *	Attribute CEHCorrectionStrength related methods
 *
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
	virtual void read_CEHCorrectionStrength(Tango::Attribute &attr);
	virtual void write_CEHCorrectionStrength(Tango::WAttribute &attr);
	virtual bool is_CEHCorrectionStrength_allowed(Tango::AttReqType type);
/**
 *	Attribute CEHResonanceStrength related methods
 *
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
	virtual void read_CEHResonanceStrength(Tango::Attribute &attr);
	virtual void write_CEHResonanceStrength(Tango::WAttribute &attr);
	virtual bool is_CEHResonanceStrength_allowed(Tango::AttReqType type);
/**
 *	Attribute CEVStrength related methods
 *	Description: embbeded corrector vertical kick
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
	virtual void read_CEVStrength(Tango::Attribute &attr);
	virtual void write_CEVStrength(Tango::WAttribute &attr);
	virtual bool is_CEVStrength_allowed(Tango::AttReqType type);
/**
 *	Attribute CEVCorrectionStrength related methods
 *
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
	virtual void read_CEVCorrectionStrength(Tango::Attribute &attr);
	virtual void write_CEVCorrectionStrength(Tango::WAttribute &attr);
	virtual bool is_CEVCorrectionStrength_allowed(Tango::AttReqType type);
/**
 *	Attribute CEVResonanceStrength related methods
 *
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
	virtual void read_CEVResonanceStrength(Tango::Attribute &attr);
	virtual void write_CEVResonanceStrength(Tango::WAttribute &attr);
	virtual bool is_CEVResonanceStrength_allowed(Tango::AttReqType type);
/**
 *	Attribute SQStrength related methods
 *	Description: embbeded skew quadrupole strength
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
	virtual void read_SQStrength(Tango::Attribute &attr);
	virtual void write_SQStrength(Tango::WAttribute &attr);
	virtual bool is_SQStrength_allowed(Tango::AttReqType type);
/**
 *	Attribute SQCorrectionStrength related methods
 *
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
	virtual void read_SQCorrectionStrength(Tango::Attribute &attr);
	virtual void write_SQCorrectionStrength(Tango::WAttribute &attr);
	virtual bool is_SQCorrectionStrength_allowed(Tango::AttReqType type);
/**
 *	Attribute SQResonanceStrength related methods
 *
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
	virtual void read_SQResonanceStrength(Tango::Attribute &attr);
	virtual void write_SQResonanceStrength(Tango::WAttribute &attr);
	virtual bool is_SQResonanceStrength_allowed(Tango::AttReqType type);
/**
 *	Attribute CEHCurrent related methods
 *
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
	virtual void read_CEHCurrent(Tango::Attribute &attr);
	virtual void write_CEHCurrent(Tango::WAttribute &attr);
	virtual bool is_CEHCurrent_allowed(Tango::AttReqType type);
/**
 *	Attribute CEVCurrent related methods
 *
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
	virtual void read_CEVCurrent(Tango::Attribute &attr);
	virtual void write_CEVCurrent(Tango::WAttribute &attr);
	virtual bool is_CEVCurrent_allowed(Tango::AttReqType type);
/**
 *	Attribute PolynomA related methods
 *	Description: array of normal multipole components
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Spectrum max = 3
 */
	virtual void read_PolynomA(Tango::Attribute &attr);
	virtual bool is_PolynomA_allowed(Tango::AttReqType type);
/**
 *	Attribute PolynomB related methods
 *	Description: array of skew multipole components
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Spectrum max = 3
 */
	virtual void read_PolynomB(Tango::Attribute &attr);
	virtual bool is_PolynomB_allowed(Tango::AttReqType type);


	//--------------------------------------------------------
	/**
	 *	Method     : E2Sextupole::add_dynamic_attributes()
	 *	Description: Add dynamic attributes if any.
	 */
	//--------------------------------------------------------
	void add_dynamic_attributes();




//	Command related methods
public:
	/**
	 *	Command Reset related method
	 *	Description: reset the corrector
	 *
	 */
	virtual void reset();
	virtual bool is_Reset_allowed(const CORBA::Any &any);
	/**
	 *	Command ResetResonanceStrength related method
	 *	Description: Transfer ResonanceStranght to CorrectionStrength and set ResonanceStrength to 0. Strngth is not modified
	 *
	 */
	virtual void reset_resonance_strength();
	virtual bool is_ResetResonanceStrength_allowed(const CORBA::Any &any);
	/**
	 *	Command ApplyDesignStrength related method
	 *	Description: Apply the memorized DesignStrength strenght to the magnet.
	 *
	 */
	virtual void apply_design_strength();
	virtual bool is_ApplyDesignStrength_allowed(const CORBA::Any &any);


	//--------------------------------------------------------
	/**
	 *	Method     : E2Sextupole::add_dynamic_commands()
	 *	Description: Add dynamic commands if any.
	 */
	//--------------------------------------------------------
	void add_dynamic_commands();

/*----- PROTECTED REGION ID(E2Sextupole::Additional Method prototypes) ENABLED START -----*/
/* clang-format on */
//	Additional Method prototypes
	void calc_rigidity(const double& energy);
	void set_main_current_write(const double& curr);
	void set_ceh_current_write(const double& curr);
	void set_cev_current_write(const double& curr);
	void update_values_write_main(void); //update internal calculated values when current write value is updated;
	void update_values_read_main(void); //update internal calculated values when current read value is updated;
	void update_values_write_ceh(void); //update internal calculated values when current ceh (SQ) write value is updated;
	void update_values_read_ceh(void); //update internal calculated values when current ceh (SQ) read value is updated;
	void update_values_write_cev(void); //update internal calculated values when current cev write value is updated;
	void update_values_read_cev(void); //update internal calculated values when current cev read value is updated;
	void update_values_state(void); //update internal calculated values when power supply state is updated;
	void init_strengths(void); // initialize strengths from power supply values
	void subscribe_events(void); //subcribe events to power supply
	//void delayed_subscribe(int delayedMs);
/* clang-format off */
/*----- PROTECTED REGION END -----*/	//	E2Sextupole::Additional Method prototypes
};

/*----- PROTECTED REGION ID(E2Sextupole::Additional Classes Definitions) ENABLED START -----*/
/* clang-format on */
//	Additional Classes Definitions
enum EventSource { MainPS, CeHPS, CeVPS};
class MainCurrentEventCallback : public Tango::CallBack
{
public:
	MainCurrentEventCallback(E2Sextupole*, EventSource);
	void push_event(Tango::EventData* ev);
protected:
	E2Sextupole* e2_sext;
	EventSource source;
};
/* clang-format off */
/*----- PROTECTED REGION END -----*/	//	E2Sextupole::Additional Classes Definitions

}	//	End of namespace

#endif   //	E2Sextupole_H
